import pytest

from little_api.utils import decode_jwt_token

from .conftest import BASE_URL


def test_basic_route_adding(api):
    @api.route("/home")
    def home(req, resp):
        resp.text = "YOLO"


def test_route_overlap_throw_exception(api):
    @api.route("/home")
    def home(req, resp):
        resp.text = "YOLO"

    with pytest.raises(AssertionError):

        @api.route("/home")
        def home2(req, resp):
            resp.text = "Next"


def test_test_client_can_send_request(api, client):
    response_text = "THIS IS COOL"

    @api.route("/hey")
    def cool(req, resp):
        resp.text = response_text

    assert client.get(f"{BASE_URL}/hey").text == response_text


def test_parameterized_route(api, client):
    @api.route("/{name}")
    def hello(req, resp, name):
        resp.text = f"hey {name}"

    assert client.get(f"{BASE_URL}/matthew").text == "hey matthew"
    assert client.get(f"{BASE_URL}/mary").text == "hey mary"


def test_default_404_response(client):
    response = client.get(f"{BASE_URL}/foo")
    assert response.status_code == 404
    assert response.text == "Not Found.."


def test_class_base_handler_get(api, client):
    response_text = "- this is a get request"

    @api.route("/book")
    class BookResource:
        def get(self, req, resp):
            resp.text = response_text

    assert client.get(f"{BASE_URL}/book").text == response_text


def test_client_base_handler_post(api, client):
    response_text = "- this is a post request"

    @api.route("/book")
    class BookResource:
        def post(self, req, resp):
            resp.text = response_text

    assert client.post(f"{BASE_URL}/book").text == response_text


def test_class_based_handler_not_allowed_method(api, client):
    @api.route("/book")
    class BookResource:
        def post(self, req, resp):
            resp.text = "yolo"

    with pytest.raises(AttributeError):
        client.get(f"{BASE_URL}/book")


def test_add_route_func(api, client):
    response_text = "Alternative way to add a route"

    def home(req, resp):
        resp.text = response_text

    api.add_route("/alternative", home)
    assert client.get(f"{BASE_URL}/alternative").text == response_text


def test_allowed_methods_for_function_based_handler(api, client):
    @api.route("/home", allowed_methods=["post"])
    def home(req, resp):
        resp.text = "hello"

    with pytest.raises(AttributeError):
        client.get(f"{BASE_URL}/home")

    assert client.post(f"{BASE_URL}/home").text == "hello"


def test_jwt_login_route(api, client):
    payload = {"user": "maddie"}

    def validate_user(request):
        return payload

    api.config["SECRET"] = "my_secret"
    api.config["JWT_EXPIRE_SECONDS"] = 100
    api.enable_jwt_login(validate_user_func=validate_user, login_route="/jwt")

    response = client.post(f"{BASE_URL}/jwt")

    assert response.status_code == 200
    token = response.json()["token"]
    claims = decode_jwt_token(token, "my_secret")
    assert claims["user"] == payload["user"]



import inspect
import os
from typing import Callable, Dict, Iterator, List, Optional, Tuple, Type

from jinja2 import Environment, FileSystemLoader
from parse import parse
from requests import Session as RequestsSession
from webob import Request
from whitenoise import WhiteNoise
from wsgiadapter import WSGIAdapter as RequestsWSGIAdapter

from little_api.exceptions import RouteNotFoundException
from little_api.response import Response
from little_api.utils import generate_jwt_token

from .config import Config
from .middleware import Middleware


class API:
    def __init__(
        self, templates_dir: str = "templates", static_dir: str = "static"
    ) -> None:
        self.routes: Dict = {}
        self.exception_handlers: Dict = {}
        self.templates_env = Environment(
            loader=FileSystemLoader(os.path.abspath(templates_dir))
        )
        self.white_noise = WhiteNoise(self.wsgi_app, root=static_dir)
        self.middleware = Middleware(self)
        self.config = Config()
        self.add_exception_handler(RouteNotFoundException, self.default_404_response)
        self._before_request = lambda res, req: None
        self._after_request = lambda res, req: None

    def __call__(self, environ: Dict, start_response: Callable) -> Iterator:
        path_info = environ["PATH_INFO"]

        if path_info.startswith("/static"):
            environ["PATH_INFO"] = path_info[len("/static") :]  # noqa
            return self.white_noise(environ, start_response)

        return self.middleware(environ, start_response)

    def before_request(self, func) -> None:
        """Methods to allow user to override"""
        self._before_request = func

    def after_request(self, func) -> None:
        """Method to allow user to override"""
        self._after_request = func

    def wsgi_app(self, environ: dict, start_response: Callable) -> Iterator:
        request = Request(environ)
        response = self.handle_request(request)
        return response(environ, start_response)

    def add_middleware(self, middleware_cls: Type[Middleware]) -> None:
        self.middleware.add(middleware_cls)

    def add_exception_handler(
        self, exception_cls: Type[Exception], handler: Callable
    ) -> None:
        self.exception_handlers[exception_cls.__name__] = handler

    def add_route(
        self, path: str, handler: Callable, allowed_methods: Optional[List] = None
    ) -> None:
        """Adds routes to known lists of paths"""
        assert path not in self.routes, f"Duplicate route found: {path}"
        if allowed_methods is None:
            allowed_methods = ["get", "post", "put", "patch", "delete", "options"]
        self.routes[path] = {"handler": handler, "allowed_methods": allowed_methods}

    def route(self, path, allowed_methods=None) -> Callable:
        """Decorator for adding routes"""

        def wrapper(handler):
            self.add_route(path, handler, allowed_methods)
            return handler

        return wrapper

    def find_handler(self, request_path: str) -> Tuple:
        """ Finds handler for a given url path"""
        for path, handler_data in self.routes.items():
            parse_result = parse(path, request_path)
            if parse_result is not None:
                return handler_data, parse_result.named
        return None, None

    def handle_request(self, request: Request) -> Response:
        """Main method to handle the request"""
        response = Response()
        self._before_request(request, response)
        handler_data, kwargs = self.find_handler(request_path=request.path)
        try:
            if handler_data is not None:
                handler = handler_data["handler"]
                allowed_methods = handler_data["allowed_methods"]
                if inspect.isclass(handler):
                    # if class get method off of class for http method
                    handler = getattr(handler(), request.method.lower(), None)
                if handler is None:
                    raise AttributeError(f"Method not allow: {request.method}")
                else:
                    if request.method.lower() not in allowed_methods:
                        raise AttributeError(f"Method not allow: {request.method}")
                handler(request, response, **kwargs)
            else:
                raise RouteNotFoundException("Not found ..")
        except Exception as e:
            exception_handler = self.exception_handlers.get(type(e).__name__)
            if exception_handler is None:
                raise e
            exception_handler(request, response, e)
        self._after_request(request, response)
        return response

    def default_404_response(self, request: Request, response: Response, exc) -> None:
        """Default response for a 404.  Can/should be overridden"""
        response.status_code = 404
        response.text = "Not Found.."

    def test_session(self, base_url="http://testserver") -> RequestsSession:
        """Used for Testing"""
        session = RequestsSession()
        session.mount(prefix=base_url, adapter=RequestsWSGIAdapter(self))
        return session

    def template(self, template_name, context: Optional[Dict] = None) -> bytes:
        if context is None:
            context = {}
        return self.templates_env.get_template(template_name).render(**context).encode()

    def enable_jwt_login(
        self, validate_user_func: Callable, login_route: str = "/jwt/login"
    ):
        def jwt_login(request, response) -> None:
            claims = validate_user_func(request)
            if claims:
                token = generate_jwt_token(
                    claims, self.config["SECRET"], self.config["JWT_EXPIRE_SECONDS"]
                )
                response.json = {"token": token}
            else:
                response.status_code = 403

        self.add_route(login_route, jwt_login, ["post"])



import pytest

from .conftest import BASE_URL


def test_json_response_helper(api, client):
    @api.route("/json")
    def json_handler(req, resp):
        resp.json = {"name": "slow"}

    response = client.get(f"{BASE_URL}/json")
    json_body = response.json()

    assert response.headers["Content-Type"] == "application/json"
    assert json_body["name"] == "slow"


def test_html_reponse_helper(api, client):
    @api.route("/html")
    def html_handler(req, resp):
        resp.html = api.template(
            "index.html", context={"title": "Best Title", "name": "Best Name"}
        )

    response = client.get(f"{BASE_URL}/html")
    assert "text/html" in response.headers["Content-Type"]
    assert "Best Title" in response.text
    assert "Best Name" in response.text


def test_text_response_helper(api, client):
    response_text = "TLDR"

    @api.route("/text")
    def text_handler(req, resp):
        resp.text = response_text

    response = client.get(f"{BASE_URL}/text")

    assert "text/plain" in response.headers["Content-Type"]
    assert response.text == response_text


def test_manually_setting_body(api, client):
    @api.route("/body")
    def text_handler(req, resp):
        resp.body = b"Byte Body"
        resp.content_type = "text/plain"

    response = client.get(f"{BASE_URL}/body")

    assert "text/plain" in response.headers["Content-Type"]
    assert response.text == "Byte Body"


def test_before_response_override(api, client):
    @api.before_request
    def before_request(request, response):
        if request.headers.get("Authorization") != "Pass":
            raise ValueError("No pass")

    @api.route("/home")
    def home(req, resp):
        resp.json = {"testing": True}

    response = client.get(f"{BASE_URL}/home", headers={"Authorization": "Pass"})
    assert response.status_code == 200

    with pytest.raises(ValueError):
        client.get(f"{BASE_URL}/home")


def test_after_response(api, client):
    @api.after_request
    def after_response(request, response):
        if response.json:
            response.json = {
                "data": response.json,
                "success": response.status_code // 100 == 2,
            }

    payload = {"value": 1}

    @api.route("/home")
    def home(req, res):
        res.json = payload
        res.status_code = 200

    res = client.get(f"{BASE_URL}/home")
    assert res.json()["data"] == payload
    assert res.json()["success"] is True



from little_api.middleware import Middleware

from .conftest import BASE_URL


def test_middleware_methods_are_called(api, client):
    process_request_called = False
    process_response_called = False

    class CallMiddlewareMethods(Middleware):
        def __init__(self, api):
            super().__init__(api)

        def process_request(self, req):
            nonlocal process_request_called
            process_request_called = True

        def process_response(self, req, resp):
            nonlocal process_response_called
            process_response_called = True

    api.add_middleware(CallMiddlewareMethods)

    @api.route("/")
    def index(res, resp):
        res.text = "YOLO"

    client.get(f"{BASE_URL}/")
    assert process_response_called is True
    assert process_request_called is True



from collections.abc import MutableMapping


class Config(MutableMapping):
    def __init__(self):
        self._config = {}

    def __setitem__(self, key, value):
        self._config[key] = value

    def __getitem__(self, key):
        return self._config[key]

    def __delitem__(self, key):
        del self._config[key]

    def __len__(self):
        return self._config.__len__()

    def __iter__(self):
        yield self._config



from .conftest import BASE_URL


def test_template(api, client):
    @api.route("/html")
    def html_handler(req, resp):
        resp.body = api.template(
            "index.html", context={"title": "Some Title", "name": "Some Name"}
        )

    response = client.get(f"{BASE_URL}/html")

    assert "text/html" in response.headers["Content-Type"]
    assert "Some Title" in response.text
    assert "Some Name" in response.text



from datetime import datetime, timedelta
from typing import Dict, List, Optional

import jwt


def generate_jwt_token(
    custom_claims: Dict, secret: str, expire_seconds: int, algorithm: str = "HS256"
):
    expire_datetime = datetime.utcnow() + timedelta(seconds=expire_seconds)
    claims = {"exp": expire_datetime}
    claims.update(custom_claims)
    token = jwt.encode(claims, secret, algorithm=algorithm)
    return token


def decode_jwt_token(
    token: str, secret: str, algorithm: Optional[List[str]] = None
) -> Dict:
    if not algorithm:
        algorithm = ["HS256"]
    claims = jwt.decode(token, secret, algorithm)  # type: ignore
    return claims



from gunicorn.app.base import BaseApplication


class DebugServer(BaseApplication):
    """Class to run Gunicorn Server in a python process for debugging"""

    def init(self, parser, opts, args):
        pass

    def __init__(
        self,
        application,
        server: str = "127.0.0.1",
        port: int = 8000,
        workers=1,
        timeout: int = 1000,
    ):
        self.options = {
            "bind": f"{server}:{port}",
            "workers": workers,
            "timeout": timeout,
        }
        self.application = application
        super().__init__()

    def load_config(self):
        config = {
            key: value
            for key, value in self.options.items()
            if key in self.cfg.settings and value is not None
        }
        for key, value in config.items():
            self.cfg.set(key.lower(), value)

    def load(self):
        return self.application



from little_api.api import API

from .conftest import BASE_URL

FILE_DIR = "css"
FILE_NAME = "main.css"
FILE_CONTENT = "body {background-color: red}"


def _create_static(static_dir):
    asset = static_dir.mkdir(FILE_DIR).join(FILE_NAME)
    asset.write(FILE_CONTENT)
    return asset


def test_404_is_returned_for_missing_static_file(client):
    assert client.get(f"{BASE_URL}/static/test.css").status_code == 404


def test_assets_are_served(tmpdir_factory):
    static_dir = tmpdir_factory.mktemp("static")
    _create_static(static_dir)
    api = API(static_dir=str(static_dir))
    client = api.test_session()

    response = client.get(f"{BASE_URL}/static/{FILE_DIR}/{FILE_NAME}")

    assert response.status_code == 200
    assert response.text == FILE_CONTENT



from little_api.exceptions import RouteNotFoundException

from .conftest import BASE_URL


def test_attribute_exception_handler(api, client):
    def custom_exception(req, resp, exc):
        resp.text = "AttributeErrorHappened"

    api.add_exception_handler(AttributeError, custom_exception)

    @api.route("/")
    def index(req, resp):
        raise AttributeError()

    response = client.get(f"{BASE_URL}/")
    assert response.text == "AttributeErrorHappened"


def test_404_exception_handler(api, client):
    def custom_404_handler(req, resp, exec):
        resp.text = "Caught 404"
        resp.status_code = 413

    api.add_exception_handler(RouteNotFoundException, custom_404_handler)

    response = client.get(f"{BASE_URL}/unknown")
    assert response.text == "Caught 404"
    assert response.status_code == 413
